from django.contrib import admin

from .models import (
    CharacterContact,
    CorporationContact,
    EveEntity,
    StandingRequest,
    StandingRevocation,
)


class AbstractStandingsRequestAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "_contact_type_str",
        "_contact_name",
        "_user",
        "request_date",
        "action_by",
        "action_date",
        "is_effective",
        "effective_date",
    )
    list_filter = ("is_effective",)
    list_select_related = True
    ordering = ("-id",)

    def _contact_name(self, obj):
        return EveEntity.objects.get_name(obj.contact_id)

    def _contact_type_str(self, obj):
        if obj.contact_type_id in CharacterContact.contact_type_ids:
            return "Character"
        elif obj.contact_type_id in CorporationContact.contact_type_ids:
            return "Corporation"
        else:
            return "(undefined)"

    _contact_type_str.short_description = "contact type"

    def _user(self, obj):
        try:
            return obj.user
        except AttributeError:
            return None

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(StandingRequest)
class StandingsRequestAdmin(AbstractStandingsRequestAdmin):
    pass


@admin.register(StandingRevocation)
class StandingsRevocationAdmin(AbstractStandingsRequestAdmin):
    pass



from allianceauth.eveonline.evelinks import eveimageserver

from ..models import CharacterAssociation, EveEntity


class EveCharacterHelper:
    """
    Mimics Alliance Auths EveCharacter with internal standingstool data instead
    """

    # Not implemented
    corporation_ticker = None
    api_id = ""
    user = None

    def __init__(self, character_id):
        self.character_id = int(character_id)
        self.main_character = None
        self.alliance_name = None
        try:
            assoc = CharacterAssociation.objects.get(character_id=self.character_id)

        except CharacterAssociation.DoesNotExist:
            self.corporation_id = None
            self.corporation_name = None
            self.alliance_id = None

        else:
            self.corporation_id = assoc.corporation_id
            self.alliance_id = assoc.alliance_id

            # Add a main character attribute (deviates from original model)
            if (
                assoc.main_character_id is not None
                and assoc.main_character_id != self.character_id
            ):
                self.main_character = EveCharacterHelper(assoc.main_character_id)

        entity_ids = [
            entity_id
            for entity_id in [self.character_id, self.corporation_id, self.alliance_id]
            if entity_id is not None
        ]
        names = EveEntity.objects.get_names(entity_ids)
        names_ids = names.keys()
        self.character_name = (
            names[self.character_id]
            if self.character_id and self.character_id in names_ids
            else None
        )
        self.corporation_name = (
            names[self.corporation_id]
            if self.corporation_id and self.corporation_id in names_ids
            else None
        )
        self.alliance_name = (
            names[self.alliance_id]
            if self.alliance_id and self.alliance_id in names_ids
            else None
        )

    def portrait_url(self, size: int = eveimageserver._DEFAULT_IMAGE_SIZE) -> str:
        return eveimageserver.character_portrait_url(self.character_id, size)



# Generated by Django 2.2.13 on 2020-07-08 12:47

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("standingsrequests", "0005_refactor_model_names"),
    ]

    operations = [
        migrations.AddField(
            model_name="standingrevocation",
            name="user",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="abstractstandingsrequest",
            name="action_date",
            field=models.DateTimeField(
                db_index=True,
                help_text="datetime of action by standing manager",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="abstractstandingsrequest",
            name="is_effective",
            field=models.BooleanField(
                db_index=True,
                default=False,
                help_text="True, when this standing is also set in-game, else False",
            ),
        ),
        migrations.AlterField(
            model_name="alliancecontact",
            name="name",
            field=models.CharField(db_index=True, max_length=254),
        ),
        migrations.AlterField(
            model_name="charactercontact",
            name="name",
            field=models.CharField(db_index=True, max_length=254),
        ),
        migrations.AlterField(
            model_name="contactlabel",
            name="name",
            field=models.CharField(db_index=True, max_length=254),
        ),
        migrations.AlterField(
            model_name="corporationcontact",
            name="name",
            field=models.CharField(db_index=True, max_length=254),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-10-06 00:47
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("standingsrequests", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="contactlabel",
            name="labelID",
            field=models.BigIntegerField(),
        ),
    ]



from bravado.exception import HTTPError

from django.contrib.auth.models import User
from django.db import models, transaction
from django.db.models import Case, Q, Value, When
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from esi.models import Token

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.eveonline.providers import ObjectNotFound
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    SR_NOTIFICATIONS_ENABLED,
    SR_OPERATION_MODE,
    STANDINGS_API_CHARID,
)
from .helpers.esi_fetch import esi_fetch

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class ContactSetManager(models.Manager):
    def create_new_from_api(self) -> object:
        """fetches contacts with standings for configured alliance
        or corporation from ESI and stores them as newly created ContactSet

        Returns new ContactSet on success, else None
        """
        token = (
            Token.objects.filter(character_id=STANDINGS_API_CHARID)
            .require_scopes(self.model.required_esi_scope())
            .require_valid()
            .first()
        )
        if not token:
            logger.warning("Token for standing char could not be found")
            return None
        try:
            contacts = _ContactsWrapper(token, STANDINGS_API_CHARID)
        except HTTPError as ex:
            logger.exception(
                "APIError occurred while trying to query api server: %s", ex
            )
            return None

        with transaction.atomic():
            contacts_set = self.create()
            self._add_labels_from_api(contacts_set, contacts.allianceLabels)
            self._add_contacts_from_api(contacts_set, contacts.alliance)

        return contacts_set

    def _add_labels_from_api(self, contact_set, labels):
        """Add the list of labels to the given ContactSet

        contact_set: ContactSet instance
        labels: Label dictionary
        """
        from .models import ContactLabel

        contact_labels = [
            ContactLabel(label_id=label.id, name=label.name, contact_set=contact_set)
            for label in labels
        ]
        ContactLabel.objects.bulk_create(contact_labels, ignore_conflicts=True)

    def _add_contacts_from_api(self, contact_set, contacts):
        """Add all contacts to the given ContactSet
        Labels _MUST_ be added before adding contacts

        :param contact_set: Django ContactSet to add contacts to
        :param contacts: List of _ContactsWrapper.Contact to add
        """
        for contact in contacts:
            flat_labels = [label.id for label in contact.labels]
            labels = contact_set.contactlabel_set.filter(label_id__in=flat_labels)
            contact_set.create_contact(
                contact_type_id=contact.type_id,
                contact_id=contact.id,
                name=contact.name,
                standing=contact.standing,
                labels=labels,
            )


class _ContactsWrapper:
    """Converts raw contacts and contact labels data from ESI into an object"""

    class Label:
        def __init__(self, json):
            self.id = json["label_id"]
            self.name = json["label_name"]

        def __str__(self):
            return u"{}".format(self.name)

        def __repr__(self):
            return str(self)

    class Contact:
        @staticmethod
        def get_type_id_from_name(type_name):
            """
            Maps new ESI name to old type id.
            Character type is allways mapped to 1373
            And faction type to 500000
            Determines the contact type:
            2 = Corporation
            1373-1386 = Character
            16159 = Alliance
            500001 - 500024 = Faction
            """
            if type_name == "character":
                return 1373
            if type_name == "alliance":
                return 16159
            if type_name == "faction":
                return 500001
            if type_name == "corporation":
                return 2

            raise NotImplementedError("This contact type is not mapped")

        def __init__(self, json, labels, names_info):
            self.id = json["contact_id"]
            self.name = names_info[self.id] if self.id in names_info else ""
            self.standing = json["standing"]
            self.in_watchlist = json["in_watchlist"] if "in_watchlist" in json else None
            self.label_ids = (
                json["label_ids"]
                if "label_ids" in json and json["label_ids"] is not None
                else []
            )
            self.type_id = self.__class__.get_type_id_from_name(json["contact_type"])
            # list of labels
            self.labels = [label for label in labels if label.id in self.label_ids]

        def __str__(self):
            return u"{}".format(self.name)

        def __repr__(self):
            return str(self)

    def __init__(self, token, character_id):
        from .models import EveEntity

        self.alliance = []
        self.allianceLabels = []

        if SR_OPERATION_MODE == "alliance":
            alliance_id = EveCharacter.objects.get_character_by_id(
                character_id
            ).alliance_id
            labels = esi_fetch(
                "Contacts.get_alliances_alliance_id_contacts_labels",
                args={"alliance_id": alliance_id},
                token=token,
            )
            for label in labels:
                self.allianceLabels.append(self.Label(label))

            contacts = esi_fetch(
                "Contacts.get_alliances_alliance_id_contacts",
                args={"alliance_id": alliance_id},
                token=token,
                has_pages=True,
            )
        elif SR_OPERATION_MODE == "corporation":
            corporation_id = EveCharacter.objects.get_character_by_id(
                character_id
            ).corporation_id
            labels = esi_fetch(
                "Contacts.get_corporations_corporation_id_contacts_labels",
                args={"corporation_id": corporation_id},
                token=token,
            )
            for label in labels:
                self.allianceLabels.append(self.Label(label))

            contacts = esi_fetch(
                "Contacts.get_corporations_corporation_id_contacts",
                args={"corporation_id": corporation_id},
                token=token,
                has_pages=True,
            )
        else:
            raise NotImplementedError()

        logger.debug("Got %d contacts in total", len(contacts))
        entity_ids = []
        for contact in contacts:
            entity_ids.append(contact["contact_id"])

        name_info = EveEntity.objects.get_names(entity_ids)
        for contact in contacts:
            self.alliance.append(self.Contact(contact, self.allianceLabels, name_info))


class AbstractStandingsRequestQuerySet(models.QuerySet):
    def annotate_is_pending(self) -> models.QuerySet:
        return self.annotate(
            is_pending_annotated=Case(
                When(Q(action_date__isnull=True) & Q(is_effective=False), then=True),
                default=Value(False),
                output_field=models.BooleanField(),
            )
        )

    def annotate_is_actioned(self) -> models.QuerySet:
        return self.annotate(
            is_actioned_annotated=Case(
                When(Q(action_date__isnull=False) & Q(is_effective=False), then=True),
                default=Value(False),
                output_field=models.BooleanField(),
            )
        )


class AbstractStandingsRequestManager(models.Manager):
    def get_queryset(self) -> models.QuerySet:
        return AbstractStandingsRequestQuerySet(self.model, using=self._db)

    def process_requests(self) -> None:
        """Process all the Standing requests/revocation objects"""
        from .models import (
            AbstractStandingsRequest,
            ContactSet,
            EveEntity,
            StandingRequest,
            StandingRevocation,
        )

        if self.model == AbstractStandingsRequest:
            raise TypeError("Can not be called from abstract objects")

        organization = ContactSet.standings_source_entity()
        organization_name = organization.name if organization else ""
        for standing_request in self.all():
            contact = EveEntity.objects.get_object(standing_request.contact_id)
            is_currently_effective = standing_request.is_effective
            is_satisfied_standing = standing_request.evaluate_effective_standing()
            if is_satisfied_standing and not is_currently_effective:
                if SR_NOTIFICATIONS_ENABLED:
                    # send notification to user about standing change if enabled
                    if type(standing_request) == StandingRequest:
                        notify(
                            user=standing_request.user,
                            title=_(
                                "%s: Standing with %s now in effect"
                                % (__title__, contact.name)
                            ),
                            message=_(
                                "'%(organization_name)s' now has blue standing with "
                                "your alt %(contact_category)s '%(contact_name)s'. "
                                "Please also update the standing of "
                                "your %(contact_category)s accordingly."
                            )
                            % {
                                "organization_name": organization_name,
                                "contact_category": contact.category,
                                "contact_name": contact.name,
                            },
                        )
                    elif type(standing_request) == StandingRevocation:
                        if standing_request.user:
                            notify(
                                user=standing_request.user,
                                title="%s: Standing with %s revoked"
                                % (__title__, contact.name),
                                message=_(
                                    "'%(organization_name)s' no longer has "
                                    "standing with your "
                                    "%(contact_category)s '%(contact_name)s'. "
                                    "Please also update the standing of "
                                    "your %(contact_category)s accordingly."
                                )
                                % {
                                    "organization_name": organization_name,
                                    "contact_category": contact.category,
                                    "contact_name": contact.name,
                                },
                            )

                # if this was a revocation the standing requests need to be remove
                # to indicate that this character no longer has standing
                if type(standing_request) == StandingRevocation:
                    StandingRequest.objects.filter(
                        contact_id=standing_request.contact_id
                    ).delete()
                    StandingRevocation.objects.filter(
                        contact_id=standing_request.contact_id
                    ).delete()

            elif is_satisfied_standing:
                # Just catching all other contact types (corps/alliances)
                # that are set effective
                pass

            elif not is_satisfied_standing and is_currently_effective:
                # Effective standing no longer effective
                logger.info(
                    "Standing for %d is marked as effective but is not "
                    "satisfied in game. Deleting." % standing_request.contact_id
                )
                standing_request.delete()

            else:
                # Check the standing hasn't been set actioned
                # and not updated in game
                actioned_timeout = standing_request.check_actioned_timeout()
                if actioned_timeout is not None and actioned_timeout:
                    logger.info(
                        "Standing request for contact ID %d has timedout "
                        "and will be reset" % standing_request.contact_id
                    )
                    if SR_NOTIFICATIONS_ENABLED:
                        title = _("Standing Request for %s reset" % contact.name)
                        message = _(
                            "The standing request for %(contact_category)s "
                            "'%(contact_name)s' from %(user_name)s "
                            "has been reset as it did not appear in "
                            "game before the timeout period expired."
                            % {
                                "contact_category": contact.category,
                                "contact_name": contact.name,
                                "user_name": standing_request.user.username,
                            },
                        )
                        # Notify standing manager
                        notify(user=actioned_timeout, title=title, message=message)
                        # Notify the user
                        notify(user=standing_request.user, title=title, message=message)

    def has_pending_request(self, contact_id: int) -> bool:
        """Checks if a request is pending for the given contact_id

        contact_id: int contact_id to check the pending request for

        returns True if a request is already pending, False otherwise
        """
        return self.pending_requests().filter(contact_id=contact_id).exists()

    def has_actioned_request(self, contact_id: int) -> bool:
        """Checks if an actioned request is pending API confirmation for
        the given contact_id

        contact_id: int contact_id to check the pending request for

        returns True if a request is pending API confirmation, False otherwise
        """
        return self.filter(
            contact_id=contact_id, action_date__isnull=False, is_effective=False
        ).exists()

    def has_effective_request(self, contact_id: int) -> bool:
        """return True if an effective request exists for given contact_id,
        else False
        """
        return self.filter(contact_id=contact_id, is_effective=True).exists()

    def pending_requests(self) -> models.QuerySet:
        """returns all pending requests for this class"""
        return self.filter(action_date__isnull=True, is_effective=False)


class StandingRequestManager(AbstractStandingsRequestManager):
    def delete_for_user(self, user):
        self.filter(user=user).delete()

    def validate_requests(self) -> int:
        """Validate all StandingsRequests and check
        that the user requesting them has permission and has API keys
        associated with the character/corp.

        StandingRevocation are created for invalid standing requests

        returns the number of invalid requests
        """
        from .models import CorporationContact, StandingRevocation

        logger.debug("Validating standings requests")
        invalid_count = 0
        for standing_request in self.all():
            logger.debug(
                "Checking request for contact_id %d", standing_request.contact_id
            )
            if not standing_request.user.has_perm(self.model.REQUEST_PERMISSION_NAME):
                logger.debug("Request is invalid, user does not have permission")
                is_valid = False

            elif CorporationContact.is_corporation(
                standing_request.contact_type_id
            ) and not self.model.can_request_corporation_standing(
                standing_request.contact_id, standing_request.user
            ):
                logger.debug("Request is invalid, not all corp API keys recorded.")
                is_valid = False

            else:
                is_valid = True

            if not is_valid:
                logger.info(
                    "Standing request for contact_id %d no longer valid. "
                    "Creating revocation",
                    standing_request.contact_id,
                )
                StandingRevocation.objects.add_revocation(
                    contact_id=standing_request.contact_id,
                    contact_type=self.model.contact_id_2_type(
                        standing_request.contact_type_id
                    ),
                    user=standing_request.user,
                )
                invalid_count += 1

        return invalid_count

    def add_request(self, user: User, contact_id: int, contact_type: str) -> object:
        """Add a new standings request

        Params:
        - user: User the request and contact_id belongs to
        - contact_id: contact_id to request standings on
        - contact_type: type of this contact

        Restuns the created StandingRequest instance
        """
        logger.debug(
            "Adding new standings request for user %s, contact %d type %s",
            user,
            contact_id,
            contact_type,
        )
        contact_type_id = self.model.contact_type_2_id(contact_type)
        if self.filter(contact_id=contact_id, contact_type_id=contact_type_id).exists():
            logger.debug(
                "Standings request already exists, " "returning first existing request"
            )
            return self.filter(contact_id=contact_id, contact_type_id=contact_type_id)[
                0
            ]

        instance = self.create(
            user=user, contact_id=contact_id, contact_type_id=contact_type_id
        )
        return instance

    def remove_requests(self, contact_id: int):
        """
        Remove the requests for the given contact_id. If any of these requests
        have been actioned or are effective
        a Revocation request will automatically be generated

        Params:
        - contact_id: contact_id to remove
        - user_responsible: User responsible for removing.
        When provided will sent notification to requestor.
        """
        logger.debug("Removing requests for contact_id %d", contact_id)
        standing_requests = self.filter(contact_id=contact_id)
        if standing_requests:
            logger.debug("%d requests to be removed", standing_requests.count())
            standing_requests.delete()


class StandingRevocationManager(AbstractStandingsRequestManager):
    def add_revocation(
        self, contact_id: int, contact_type: str, user: User = None
    ) -> object:
        """Add a new standings revocation

        Params:
        - contact_id: contact_id to request standings on
        - contact_type_id: contact_type_id from AbstractContact concrete implementation
        - user: user making the request

        Returns the created StandingRevocation instance
        """
        logger.debug(
            "Adding new standings revocation for contact %d type %s",
            contact_id,
            contact_type,
        )
        contact_type_id = self.model.contact_type_2_id(contact_type)
        pending = self.filter(contact_id=contact_id).filter(is_effective=False)
        if pending.exists():
            logger.debug(
                "Cannot add revocation for contact %d %s, " "pending revocation exists",
                contact_id,
                contact_type_id,
            )
            return None

        instance = self.create(
            contact_id=contact_id, contact_type_id=contact_type_id, user=user
        )
        return instance


class CharacterAssociationManager(models.Manager):
    def update_from_auth(self) -> None:
        """Update all character associations based on auth relationship data"""
        from .models import EveEntity

        for character in EveCharacter.objects.all():
            logger.debug(
                "Updating Association from Auth for %s", character.character_name
            )
            try:
                ownership = CharacterOwnership.objects.get(character=character)
            except CharacterOwnership.DoesNotExist:
                main = None
            else:
                main = (
                    ownership.user.profile.main_character.character_id
                    if ownership.user.profile.main_character
                    else None
                )

            self.update_or_create(
                character_id=character.character_id,
                defaults={
                    "corporation_id": character.corporation_id,
                    "main_character_id": main,
                    "alliance_id": character.alliance_id,
                    "updated": now(),
                },
            )
            EveEntity.objects.update_or_create_from_evecharacter(character)

    def update_from_api(self) -> None:
        """Update all character corp associations we have standings for that
        aren't being updated locally
        Cache timeout should be longer than update_from_auth
        update schedule to
        prevent unnecessarily updating characters we already have local data for.
        """
        # gather character associations of pilots which meed to be updated
        from .models import ContactSet, EveEntity

        try:
            contact_set = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            logger.warning("Could not find a contact set")
        else:
            all_pilots = contact_set.charactercontact_set.values_list(
                "contact_id", flat=True
            )
            expired_character_associations = self.get_api_expired_items(
                all_pilots
            ).values_list("character_id", flat=True)
            expired_pilots = set(all_pilots).intersection(
                expired_character_associations
            )
            known_pilots = self.values_list("character_id", flat=True)
            unknown_pilots = [
                pilot for pilot in all_pilots if pilot not in known_pilots
            ]
            pilots_to_fetch = list(expired_pilots | set(unknown_pilots))

            # Fetch the data in acceptable sizes from the API
            chunk_size = 1000
            for pilots_chunk in chunks(pilots_to_fetch, chunk_size):
                try:
                    esi_response = esi_fetch(
                        "Character.post_characters_affiliation",
                        args={"characters": pilots_chunk},
                    )
                except HTTPError:
                    logger.exception(
                        "Could not fetch associations pilots_chunk from ESI"
                    )
                else:
                    entity_ids = []
                    for association in esi_response:
                        corporation_id = association["corporation_id"]
                        alliance_id = (
                            association["alliance_id"]
                            if "alliance_id" in association
                            else None
                        )
                        character_id = association["character_id"]
                        self.update_or_create(
                            character_id=character_id,
                            defaults={
                                "corporation_id": corporation_id,
                                "alliance_id": alliance_id,
                                "updated": now(),
                            },
                        )
                        entity_ids.append(character_id)
                        entity_ids.append(corporation_id)
                        if alliance_id:
                            entity_ids.append(alliance_id)

                    # make sure we have all names resolved
                    EveEntity.objects.get_names(entity_ids)

    def get_api_expired_items(self, items_in=None) -> models.QuerySet:
        """Get all API timer expired items

        items_in: list optional parameter to limit the results
        to character_ids in the list

        returns: QuerySet of CharacterAssociation items
        """
        expired = self.filter(updated__lt=now() - self.model.API_CACHE_TIMER)
        if items_in is not None:
            expired = expired.filter(character_id__in=items_in)

        return expired


class EveEntityManager(models.Manager):
    def get_names(self, entity_ids: list) -> dict:
        """Get the names of the given entity ids from catch or other locations

        eve_entity_ids: array of int entity ids who's names to fetch

        returns dict with entity_id as key and name as value
        """
        entity_ids = set(entity_ids)  # remove duplicates
        entity_ids_known = set(
            self.filter(entity_id__in=entity_ids).values_list("entity_id", flat=True)
        )
        entity_ids_unknown = entity_ids - entity_ids_known
        entities_expired = self.filter(
            entity_id__in=entity_ids, updated__lt=now() - self.model.CACHE_TIME
        )
        entity_ids_need_update = list(entity_ids_unknown) + list(
            entities_expired.values_list("entity_id", flat=True)
        )
        if entity_ids_need_update:
            entities_api_info = self._get_names_from_api(entity_ids_need_update)

            # update existing entities
            entities_to_update = list()
            for entity in entities_expired:
                if entity.entity_id in entities_api_info:
                    entity.name = entities_api_info[entity.entity_id]["name"]
                    entity.category = entities_api_info[entity.entity_id]["category"]
                    entities_to_update.append(entity)

            if entities_to_update:
                self.bulk_update(
                    entities_to_update, fields=["name", "category"], batch_size=500
                )

            # create missing entities
            entities_to_create = list()
            for entity_id in entity_ids_unknown:
                if entity_id in entities_api_info:
                    entity = self.model(
                        entity_id=entity_id,
                        name=entities_api_info[entity_id]["name"],
                        category=entities_api_info[entity_id]["category"],
                    )
                    entities_to_create.append(entity)

            if entities_to_create:
                self.bulk_create(
                    entities_to_create, ignore_conflicts=True, batch_size=500
                )

        return {
            entity.entity_id: entity.name
            for entity in self.filter(entity_id__in=entity_ids)
        }

    @classmethod
    def _get_names_from_api(cls, eve_entity_ids) -> dict:
        """
        Get the names of the given entity ids from the EVE API servers
        :param eve_entity_ids: array of int entity ids who's names to fetch
        :return: dict with entity_id as key and name as value
        """
        # this is to make sure there are no duplicates
        eve_entity_ids = list(set(eve_entity_ids))

        entities_list = list()
        chunk_size = 1000
        for ids_chunk in chunks(eve_entity_ids, chunk_size):
            entities_list += cls._make_api_request(ids_chunk)

        return {entity["id"]: entity for entity in entities_list}

    @classmethod
    def _make_api_request(cls, eve_entity_ids: list) -> list:
        """
        Get the names of the given entity ids from the EVE API servers
        :param eve_entity_ids: array of int entity ids who's names to fetch
        :return: array of objects with keys id and name or None if unsuccessful
        """
        logger.debug(
            "Attempting to get entity name from API for ids %s", eve_entity_ids
        )
        try:
            return esi_fetch(
                "Universe.post_universe_names", args={"ids": eve_entity_ids}
            )

        except HTTPError:
            logger.exception(
                "Error occurred while trying to query api for entity id=%s",
                eve_entity_ids,
            )
            raise ObjectNotFound(eve_entity_ids, "universe_entities")

    def get_name(self, entity_id: int) -> str:
        """Get the name for the given entity

        entity_id: EVE id of the entity

        returns name if it exists or None
        """
        if not entity_id:
            return None

        names_info = self.get_names([entity_id])
        return names_info[entity_id] if entity_id in names_info else None

    def get_object(self, entity_id: int) -> object:
        self.get_name(entity_id)
        return self.get(entity_id=entity_id)

    def update_name(self, entity_id, name, category=None):
        self.update_or_create(
            entity_id=entity_id,
            defaults={
                "name": name,
                "category": category,
            },
        )

    def update_or_create_from_evecharacter(self, character: EveCharacter) -> None:
        self.update_or_create(
            entity_id=character.character_id,
            defaults={
                "name": character.character_name,
                "category": self.model.CATEGORY_CHARACTER,
            },
        )
        self.update_or_create(
            entity_id=character.corporation_id,
            defaults={
                "name": character.corporation_name,
                "category": self.model.CATEGORY_CORPORATION,
            },
        )
        if character.alliance_id:
            self.update_or_create(
                entity_id=character.alliance_id,
                defaults={
                    "name": character.alliance_name,
                    "category": self.model.CATEGORY_ALLIANCE,
                },
            )



from functools import wraps

from esi.decorators import _check_callback
from esi.models import Token
from esi.views import select_token, sso_redirect

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .models import StandingRequest

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def token_required_by_state(new=False):
    """
    Decorator for views which supplies a single,
    user-selected token for the view to process.
    No scopes can be provided instead scopes are selected
    from the scopes that are set as
    required for a specific state in the settings.
    param: new if a new token should be aquired
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            scopes = ""
            if request.user.profile.state is not None:
                scopes = " ".join(
                    StandingRequest.get_required_scopes_for_state(
                        request.user.profile.state.name
                    )
                )

            # if we're coming back from SSO for a new token, return it
            token = _check_callback(request)
            if token and new:
                logger.debug(
                    "Got new token from %s session %s. Returning to view.",
                    request.user,
                    request.session.session_key[:5],
                )
                return view_func(request, token, *args, **kwargs)

            # if we're selecting a token, return it
            if request.method == "POST":
                if request.POST.get("_add", False):
                    logger.debug(
                        "%s has selected to add new token. Redirecting to SSO.",
                        request.user,
                    )
                    # user has selected to add a new token
                    return sso_redirect(request, scopes=scopes)

                token_pk = request.POST.get("_token", None)
                if token_pk:
                    token_pk = int(token_pk)
                    logger.debug("%s has selected token %d", request.user, token_pk)
                    try:
                        token = Token.objects.get(pk=token_pk)
                        # ensure token belongs to this user and has required scopes
                        if (
                            (token.user and token.user == request.user)
                            or not token.user
                        ) and Token.objects.filter(pk=token_pk).require_scopes(
                            scopes
                        ).require_valid().exists():
                            logger.debug(
                                "Selected token fulfills requirements of view."
                                " Returning."
                            )
                            return view_func(request, token, *args, **kwargs)
                    except Token.DoesNotExist:
                        logger.debug("Token %d not found.", token_pk)

            if not new:
                # present the user with token choices
                tokens = (
                    Token.objects.filter(user__pk=request.user.pk)
                    .require_scopes(scopes)
                    .require_valid()
                )
                if tokens.exists():
                    logger.debug(
                        "Returning list of available tokens for %s.", request.user
                    )
                    return select_token(request, scopes=scopes, new=new)
                else:
                    logger.debug(
                        "No tokens found for %s session %s with scopes %s",
                        request.user,
                        request.session.session_key[:5],
                        scopes,
                    )

            # prompt the user to add a new token
            logger.debug(
                "Redirecting %s session %s to SSO.",
                request.user,
                request.session.session_key[:5],
            )
            return sso_redirect(request, scopes=scopes)

        return _wrapped_view

    return decorator



from datetime import timedelta

from django.contrib.auth.models import User
from django.core import exceptions
from django.db import models
from django.utils.timezone import now
from esi.models import Token

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.evelinks import eveimageserver
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    SR_OPERATION_MODE,
    SR_REQUIRED_SCOPES,
    SR_STANDING_TIMEOUT_HOURS,
    STANDINGS_API_CHARID,
    STR_ALLIANCE_IDS,
    STR_CORP_IDS,
)
from .helpers.evecorporation import EveCorporation
from .managers import (
    AbstractStandingsRequestManager,
    CharacterAssociationManager,
    ContactSetManager,
    EveEntityManager,
    StandingRequestManager,
    StandingRevocationManager,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class ContactSet(models.Model):
    """Set of contacts from configured alliance or corporation
    which defines its current standings
    """

    date = models.DateTimeField(auto_now_add=True, db_index=True)
    name = models.CharField(max_length=254)

    objects = ContactSetManager()

    class Meta:
        get_latest_by = "date"
        permissions = (
            ("view", "User can view standings"),
            ("download", "User can export standings to a CSV file"),
        )

    def __str__(self):
        return str(self.date)

    def __repr__(self):
        return f"{type(self).__name__}(pk={self.pk}, date='{self.date}')"

    def get_contact_by_id(self, contact_id: int, contact_type_id: int) -> object:
        """Attempts to fetch the contact for the given ID and type

        Params:
        - contact_id: Integer contact ID
        - contact_type_id: Integer contact type from the contact_type_ids attribute
        in concrete models

        Returns concrete contact Object or ObjectDoesNotExist exception
        """
        if contact_type_id in CharacterContact.contact_type_ids:
            return self.charactercontact_set.get(contact_id=contact_id)
        elif contact_type_id in CorporationContact.contact_type_ids:
            return self.corporationcontact_set.get(contact_id=contact_id)
        elif contact_type_id in AllianceContact.contact_type_ids:
            return self.alliancecontact_set.get(contact_id=contact_id)
        raise exceptions.ObjectDoesNotExist()

    def create_contact(
        self,
        contact_type_id: int,
        contact_id: int,
        name: str,
        standing: float,
        labels: list,
    ) -> object:
        """creates new contact"""
        StandingType = self.get_class_for_contact_type(contact_type_id)
        contact = StandingType.objects.create(
            contact_set=self,
            contact_id=contact_id,
            name=name,
            standing=standing,
        )
        for label in labels:
            contact.labels.add(label)

        return contact

    def character_has_satisfied_standing(self, contact_id: int) -> bool:
        return self.contact_has_satisfied_standing(
            contact_id, CharacterContact.get_contact_type_id()
        )

    def corporation_has_satisfied_standing(self, contact_id: int) -> bool:
        return self.contact_has_satisfied_standing(
            contact_id, CorporationContact.get_contact_type_id()
        )

    def contact_has_satisfied_standing(
        self, contact_id: int, contact_type_id: int
    ) -> bool:
        """Return True if give contact has standing exists"""
        if contact_type_id in CharacterContact.contact_type_ids:
            try:
                contact = self.charactercontact_set.get(contact_id=contact_id)
            except CharacterContact.DoesNotExist:
                return False

        elif contact_type_id in CorporationContact.contact_type_ids:
            try:
                contact = self.corporationcontact_set.get(contact_id=contact_id)
            except CorporationContact.DoesNotExist:
                return False

        elif contact_type_id in AllianceContact.contact_type_ids:
            try:
                contact = self.alliancecontact_set.get(contact_id=contact_id)
            except CorporationContact.DoesNotExist:
                return False

        else:
            raise ValueError("Invalid contact type ID: %s" % contact_type_id)

        return StandingRequest.is_standing_satisfied(contact.standing)

    @staticmethod
    def get_class_for_contact_type(contact_type_id):
        if contact_type_id in CharacterContact.contact_type_ids:
            return CharacterContact
        elif contact_type_id in CorporationContact.contact_type_ids:
            return CorporationContact
        elif contact_type_id in AllianceContact.contact_type_ids:
            return AllianceContact
        raise ValueError("Invalid contact type ID: %s" % contact_type_id)

    @classmethod
    def is_character_in_organisation(cls, character: EveCharacter) -> bool:
        """Check if the Pilot is in the auth instances organisation

        character: EveCharacter

        returns True if the character is in the organisation, False otherwise
        """
        return (
            character.corporation_id in cls.corporation_ids_in_organization()
            or character.alliance_id in cls.alliance_ids_in_organization()
        )

    def corporation_ids_in_organization() -> set:
        return {int(corporation_id) for corporation_id in list(STR_CORP_IDS)}

    @staticmethod
    def alliance_ids_in_organization() -> set:
        return {int(corporation_id) for corporation_id in list(STR_ALLIANCE_IDS)}

    @staticmethod
    def required_esi_scope() -> str:
        """returns the required ESI scopes for syncing"""
        if SR_OPERATION_MODE == "alliance":
            return "esi-alliances.read_contacts.v1"
        elif SR_OPERATION_MODE == "corporation":
            return "esi-corporations.read_contacts.v1"
        else:
            raise NotImplementedError()

    @staticmethod
    def standings_character() -> EveCharacter:
        """returns the configured standings character"""
        try:
            character = EveCharacter.objects.get(character_id=STANDINGS_API_CHARID)
        except EveCharacter.DoesNotExist:
            character = EveCharacter.objects.create_character(STANDINGS_API_CHARID)
            EveEntity.objects.get_or_create(
                entity_id=character.character_id,
                defaults={
                    "name": character.character_name,
                    "category": EveEntity.CATEGORY_CHARACTER,
                },
            )

        return character

    @classmethod
    def standings_source_entity(cls) -> object:
        """returns the entity that all standings are fetched from

        returns None when in alliance mode, but character has no alliance
        """
        character = cls.standings_character()
        if SR_OPERATION_MODE == "alliance":
            if character.alliance_id:
                entity, _ = EveEntity.objects.get_or_create(
                    entity_id=character.alliance_id,
                    defaults={
                        "name": character.alliance_name,
                        "category": EveEntity.CATEGORY_ALLIANCE,
                    },
                )
            else:
                entity = None
        elif SR_OPERATION_MODE == "corporation":
            entity, _ = EveEntity.objects.get_or_create(
                entity_id=character.corporation_id,
                defaults={
                    "name": character.corporation_name,
                    "category": EveEntity.CATEGORY_CORPORATION,
                },
            )
        else:
            raise NotImplementedError()

        return entity

    def generate_standing_requests_for_blue_alts(self) -> int:
        """Automatically creates effective standings requests for
        alt characters on Auth that already have blue standing in-game.

        return count of generated standings requests
        """
        logger.info("Started generating standings request for blue alts.")
        owned_characters_qs = EveCharacter.objects.filter(
            character_ownership__isnull=False
        ).select_related()
        created_counter = 0
        for alt in owned_characters_qs:
            user = alt.character_ownership.user
            if (
                not self.is_character_in_organisation(alt)
                and not StandingRequest.objects.filter(
                    user=user, contact_id=alt.character_id
                ).exists()
                and not StandingRevocation.objects.filter(
                    contact_id=alt.character_id
                ).exists()
                and self.character_has_satisfied_standing(alt.character_id)
            ):
                sr = StandingRequest.objects.add_request(
                    user=user,
                    contact_id=alt.character_id,
                    contact_type=StandingRequest.CHARACTER_CONTACT_TYPE,
                )
                sr.mark_actioned(None)
                sr.mark_effective()
                logger.info(
                    "Generated standings request for blue alt %s "
                    "belonging to user %s.",
                    alt,
                    user,
                )
                created_counter += 1

        logger.info(
            "Completed generating %d standings request for blue alts.",
            created_counter,
        )
        return created_counter


class ContactLabel(models.Model):
    """A contact label"""

    contact_set = models.ForeignKey(ContactSet, on_delete=models.CASCADE)
    label_id = models.BigIntegerField(db_index=True)
    name = models.CharField(max_length=254, db_index=True)

    def __str__(self):
        return self.name

    def __repr__(self):
        return (
            f"{type(self).__name__}(pk={self.pk}, "
            f"label_id={self.label_id}, name='{self.name}')"
        )


class AbstractContact(models.Model):
    """Base class for a contact"""

    CHARACTER_AMARR_TYPE_ID = 1373
    CHARACTER_NI_KUNNI_TYPE_ID = 1374
    CHARACTER_CIVRE_TYPE_ID = 1375
    CHARACTER_DETEIS_TYPE_ID = 1376
    CHARACTER_GALLENTE_TYPE_ID = 1377
    CHARACTER_INTAKI_TYPE_ID = 1378
    CHARACTER_SEBIESTOR_TYPE_ID = 1379
    CHARACTER_BRUTOR_TYPE_ID = 1380
    CHARACTER_STATIC_TYPE_ID = 1381
    CHARACTER_MODIFIER_TYPE_ID = 1382
    CHARACTER_ACHURA_TYPE_ID = 1383
    CHARACTER_JIN_MEI_TYPE_ID = 1384
    CHARACTER_KHANID_TYPE_ID = 1385
    CHARACTER_VHEROKIOR_TYPE_ID = 1386
    CHARACTER_DRIFTER_TYPE_ID = 34574
    ALLIANCE_TYPE_ID = 16159
    CORPORATION_TYPE_ID = 2

    contact_set = models.ForeignKey(ContactSet, on_delete=models.CASCADE)
    contact_id = models.PositiveIntegerField(db_index=True)
    name = models.CharField(max_length=254, db_index=True)
    standing = models.FloatField(db_index=True)
    labels = models.ManyToManyField(ContactLabel)

    class Meta:
        abstract = True

    def __repr__(self):
        return (
            f"{type(self).__name__}(pk={self.pk}, "
            f"contact_id={self.contact_id}, name='{self.name}', "
            f"standing={self.standing})"
        )

    @classmethod
    def get_contact_type_id(cls) -> int:
        """returns the contact type ID for this type of contact"""
        try:
            return cls.contact_type_ids[0]
        except AttributeError:
            raise NotImplementedError() from None

    @classmethod
    def get_contact_type_ids(cls) -> int:
        """returns the list of valid contact type IDs for this type of contact"""
        try:
            return cls.contact_type_ids
        except AttributeError:
            raise NotImplementedError() from None

    @staticmethod
    def is_character(type_id):
        return type_id in CharacterContact.contact_type_ids

    @staticmethod
    def is_corporation(type_id):
        return type_id in CorporationContact.contact_type_ids

    @staticmethod
    def is_alliance(type_id):
        return type_id in AllianceContact.contact_type_ids


class CharacterContact(AbstractContact):
    """A character contact"""

    contact_type_ids = [
        AbstractContact.CHARACTER_AMARR_TYPE_ID,
        AbstractContact.CHARACTER_NI_KUNNI_TYPE_ID,
        AbstractContact.CHARACTER_CIVRE_TYPE_ID,
        AbstractContact.CHARACTER_DETEIS_TYPE_ID,
        AbstractContact.CHARACTER_GALLENTE_TYPE_ID,
        AbstractContact.CHARACTER_INTAKI_TYPE_ID,
        AbstractContact.CHARACTER_SEBIESTOR_TYPE_ID,
        AbstractContact.CHARACTER_BRUTOR_TYPE_ID,
        AbstractContact.CHARACTER_STATIC_TYPE_ID,
        AbstractContact.CHARACTER_MODIFIER_TYPE_ID,
        AbstractContact.CHARACTER_ACHURA_TYPE_ID,
        AbstractContact.CHARACTER_JIN_MEI_TYPE_ID,
        AbstractContact.CHARACTER_KHANID_TYPE_ID,
        AbstractContact.CHARACTER_VHEROKIOR_TYPE_ID,
        AbstractContact.CHARACTER_DRIFTER_TYPE_ID,
    ]
    is_watched = models.BooleanField(default=False)


class CorporationContact(AbstractContact):
    """A corporation contact"""

    contact_type_ids = [AbstractContact.CORPORATION_TYPE_ID]


class AllianceContact(AbstractContact):
    """An alliance contact"""

    contact_type_ids = [AbstractContact.ALLIANCE_TYPE_ID]


class AbstractStandingsRequest(models.Model):
    """Base class for a standing request"""

    # possible contact types to make a request for
    CHARACTER_CONTACT_TYPE = "character"
    CORPORATION_CONTACT_TYPE = "corporation"

    # Standing less than or equal
    EXPECT_STANDING_LTEQ = 10.0

    # Standing greater than or equal
    EXPECT_STANDING_GTEQ = -10.0

    # permission needed to request standing
    REQUEST_PERMISSION_NAME = "standingsrequests.request_standings"

    contact_id = models.PositiveIntegerField(
        db_index=True, help_text="EVE Online ID of contact this standing is for"
    )
    contact_type_id = models.PositiveIntegerField(
        db_index=True, help_text="EVE Online Type ID of this contact"
    )
    request_date = models.DateTimeField(
        auto_now_add=True, db_index=True, help_text="datetime this request was created"
    )
    action_by = models.ForeignKey(
        User,
        default=None,
        null=True,
        on_delete=models.SET_DEFAULT,
        db_index=True,
        help_text="standing manager that accepted or rejected this requests",
    )
    action_date = models.DateTimeField(
        null=True, db_index=True, help_text="datetime of action by standing manager"
    )
    is_effective = models.BooleanField(
        default=False,
        db_index=True,
        help_text="True, when this standing is also set in-game, else False",
    )
    effective_date = models.DateTimeField(
        null=True, help_text="Datetime when this standing was set active in-game"
    )

    objects = AbstractStandingsRequestManager()

    class Meta:
        permissions = (
            ("affect_standings", "User can process standings requests."),
            ("request_standings", "User can request standings."),
        )

    def __repr__(self) -> str:
        try:
            user_str = f", user='{self.user}'"
        except AttributeError:
            user_str = ""

        return (
            f"{type(self).__name__}(pk={self.pk}, contact_id={self.contact_id}"
            f"{user_str}, is_effective={self.is_effective})"
        )

    @property
    def is_character(self) -> bool:
        return AbstractContact.is_character(self.contact_type_id)

    @property
    def is_corporation(self) -> bool:
        return AbstractContact.is_corporation(self.contact_type_id)

    @property
    def is_actioned(self) -> bool:
        return self.action_date is not None and not self.is_effective

    @property
    def is_pending(self) -> bool:
        return self.action_date is None and self.is_effective is False

    @classmethod
    def is_standing_satisfied(cls, standing: float) -> bool:
        if standing is not None:
            return (
                cls.EXPECT_STANDING_GTEQ <= float(standing) <= cls.EXPECT_STANDING_LTEQ
            )
        else:
            return False

    @classmethod
    def contact_type_2_id(cls, contact_type) -> int:
        if contact_type == cls.CHARACTER_CONTACT_TYPE:
            return CharacterContact.get_contact_type_id()
        elif contact_type == cls.CORPORATION_CONTACT_TYPE:
            return CorporationContact.get_contact_type_id()
        else:
            raise ValueError("Invalid contact type")

    @classmethod
    def contact_id_2_type(cls, contact_type_id) -> str:
        if contact_type_id in CharacterContact.get_contact_type_ids():
            return cls.CHARACTER_CONTACT_TYPE
        elif contact_type_id in CorporationContact.get_contact_type_ids():
            return cls.CORPORATION_CONTACT_TYPE
        else:
            raise ValueError("Invalid contact type")

    def evaluate_effective_standing(self, check_only: bool = False) -> bool:
        """
        Check and mark a standing as satisfied
        :param check_only: Check the standing only, take no action
        """
        try:
            logger.debug("Checking standing for %d", self.contact_id)
            latest = ContactSet.objects.latest()
            contact = latest.get_contact_by_id(self.contact_id, self.contact_type_id)
            if self.is_standing_satisfied(contact.standing):
                # Standing is satisfied
                logger.debug("Standing satisfied for %d", self.contact_id)
                if not check_only:
                    self.mark_effective()
                return True

        except exceptions.ObjectDoesNotExist:
            logger.debug(
                "No standing set for %d, checking if neutral is OK", self.contact_id
            )
            if self.is_standing_satisfied(0):
                # Standing satisfied but deleted (neutral)
                logger.debug(
                    "Standing satisfied but deleted (neutral) for %d", self.contact_id
                )
                if not check_only:
                    self.mark_effective()
                return True

        # Standing not satisfied
        logger.debug("Standing NOT satisfied for %d", self.contact_id)
        return False

    def mark_effective(self, date=None):
        """
        Marks a standing as effective (standing exists in game)
        from the current or supplied TZ aware datetime
        :param date: TZ aware datetime object of when the standing became effective
        :return:
        """
        logger.debug("Marking standing for %d as effective", self.contact_id)
        self.is_effective = True
        self.effective_date = date if date else now()
        self.save()

    def mark_actioned(self, user, date=None):
        """
        Marks a standing as actioned (user has made the change in game)
        with the current or supplied TZ aware datetime
        :param user: Actioned By django User
        :param date: TZ aware datetime object of when the action was taken
        :return:
        """
        logger.debug("Marking standing for %d as actioned", self.contact_id)
        self.action_by = user
        self.action_date = date if date else now()
        self.save()

    def check_actioned_timeout(self):
        """
        Check that a standing hasn't been marked as actioned
        and is still not effective ~24hr later
        :return: User if the actioned has timed out, False if it has not,
        None if the check was unsuccessful
        """
        logger.debug("Checking standings request timeout")
        if self.is_effective:
            logger.debug("Standing is already marked as effective...")
            return None

        if self.action_by is None:
            logger.debug("Standing was never actioned, cannot timeout")
            return None

        try:
            latest = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            logger.debug("Cannot check standing timeout, no standings available")
            return None

        # Reset request that has not become effective after timeout expired
        if self.action_date + timedelta(hours=SR_STANDING_TIMEOUT_HOURS) < latest.date:
            logger.info(
                "Standing actioned timed out, resetting actioned for contact_id %d",
                self.contact_id,
            )
            actioner = self.action_by
            self.action_by = None
            self.action_date = None
            self.save()
            return actioner
        return False

    def reset_to_initial(self) -> None:
        """
        Reset a standing back to its initial creation state
        (Not actioned and not effective)
        :return:
        """
        self.is_effective = False
        self.effective_date = None
        self.action_by = None
        self.action_date = None
        self.save()


class StandingRequest(AbstractStandingsRequest):
    """A change request to get standing for a character or corporation

    OR a record representing that a character or corporation currently has standing

    Standing Requests (SR) can have one of 3 states:
    - new: Newly created SRs represent a new request from a user. They are not actioned and not effective
    - actionied: A standing manager marks a SR as actionied, once he has set the new standing in-game
    - effective: Once the new standing is returned from the API a SR is marked effective. Effective SRs stay in database to represent that a user has standing.
    """

    EXPECT_STANDING_GTEQ = 0.01

    user = models.ForeignKey(User, on_delete=models.CASCADE)

    objects = StandingRequestManager()

    def delete(self, using=None, keep_parents=False):
        """
        Add a revocation before deleting if the standing has been
        actioned (pending) or is effective and
        doesn't already have a pending revocation request.
        """
        if self.action_by is not None or self.is_effective:
            # Check if theres not already a revocation pending
            if not StandingRevocation.objects.has_pending_request(self.contact_id):
                logger.debug(
                    "Adding revocation for deleted request "
                    "with contact_id %d type %s",
                    self.contact_id,
                    self.contact_type_id,
                )
                StandingRevocation.objects.add_revocation(
                    contact_id=self.contact_id,
                    contact_type=self.contact_id_2_type(self.contact_type_id),
                    user=self.user,
                )
            else:
                logger.debug(
                    "Revocation already pending for deleted request "
                    "with contact_id %d type %s",
                    self.contact_id,
                    self.contact_type_id,
                )
        else:
            logger.debug(
                "Standing never effective, no revocation required "
                "for deleted request with contact_id %d type %s",
                self.contact_id,
                self.contact_type_id,
            )

        super().delete(using, keep_parents)

    @classmethod
    def can_request_corporation_standing(cls, corporation_id: int, user: User) -> bool:
        """
        Checks if given user owns all of the required corp tokens for standings to be permitted

        Params
        - corporation_id: corp to check for
        - user: User to check for

        returns True if they can request standings, False if they cannot
        """
        corporation = EveCorporation.get_by_id(corporation_id)
        return (
            corporation is not None
            and not corporation.is_npc
            and corporation.user_has_all_member_tokens(user)
        )

    @classmethod
    def has_required_scopes_for_request(
        cls, character: EveCharacter, user: User = None, quick_check: bool = False
    ) -> bool:
        """returns true if given character has the required scopes
        for issueing a standings request else false

        Params:
        - user: provide User object to shorten processing time
        - quick: if True will not check if tokens are valid to save time
        """
        if not user:
            try:
                ownership = CharacterOwnership.objects.select_related(
                    "user__profile__state"
                ).get(character__character_id=character.character_id)
            except CharacterOwnership.DoesNotExist:
                return False
            else:
                user = ownership.user

        state_name = user.profile.state.name
        scopes_string = " ".join(cls.get_required_scopes_for_state(state_name))
        token_qs = Token.objects.filter(
            character_id=character.character_id
        ).require_scopes(scopes_string)
        if not quick_check:
            token_qs = token_qs.require_valid()

        return token_qs.exists()

    @staticmethod
    def get_required_scopes_for_state(state_name: str) -> list:
        state_name = "" if not state_name else state_name
        return (
            SR_REQUIRED_SCOPES[state_name]
            if state_name in SR_REQUIRED_SCOPES
            else list()
        )


class StandingRevocation(AbstractStandingsRequest):
    """A standing revocation"""

    EXPECT_STANDING_LTEQ = 0.0

    user = models.ForeignKey(
        User, on_delete=models.SET_DEFAULT, default=None, null=True
    )

    objects = StandingRevocationManager()


class CharacterAssociation(models.Model):
    """
    Alt Character Associations with declared mains
    Main characters are associated with themselves
    """

    API_CACHE_TIMER = timedelta(days=3)

    character_id = models.PositiveIntegerField(primary_key=True)
    corporation_id = models.PositiveIntegerField(null=True)
    alliance_id = models.PositiveIntegerField(null=True)
    main_character_id = models.PositiveIntegerField(null=True)
    updated = models.DateTimeField(auto_now_add=True)

    objects = CharacterAssociationManager()

    @property
    def character_name(self):
        """
        Character name property for character_id
        :return: str character name
        """
        name = EveEntity.objects.get_name(self.character_id)
        return name

    @property
    def main_character_name(self):
        """
        Character name property for character_id
        :return: str character name
        """
        if self.main_character_id:
            name = EveEntity.objects.get_name(self.main_character_id)
        else:
            name = None
        return name


class EveEntity(models.Model):
    """An Eve Online entity like a character or a corporation

    A main function of this class is to enable name matching for Eve IDs
    """

    CACHE_TIME = timedelta(days=30)

    CATEGORY_ALLIANCE = "alliance"
    CATEGORY_CHARACTER = "character"
    CATEGORY_CONSTELLATION = "constellation"
    CATEGORY_CORPORATION = "corporation"
    CATEGORY_FACTION = "faction"
    CATEGORY_INVENTORY_TYPE = "inventory_type"
    CATEGORY_REGION = "region"
    CATEGORY_SOLAR_SYSTEM = "solar_system"
    CATEGORY_STATION = "station"

    CATEGORY_CHOICES = (
        (CATEGORY_ALLIANCE, "alliance"),
        (CATEGORY_CHARACTER, "character"),
        (CATEGORY_CONSTELLATION, "constellation"),
        (CATEGORY_CORPORATION, "corporation"),
        (CATEGORY_FACTION, "faction"),
        (CATEGORY_INVENTORY_TYPE, "inventory_type"),
        (CATEGORY_REGION, "region"),
        (CATEGORY_SOLAR_SYSTEM, "solar_system"),
        (CATEGORY_STATION, "station"),
    )

    entity_id = models.PositiveIntegerField(primary_key=True)
    name = models.CharField(max_length=254)
    category = models.CharField(
        max_length=16, choices=CATEGORY_CHOICES, default=None, null=True
    )
    updated = models.DateTimeField(auto_now=True, db_index=True)

    objects = EveEntityManager()

    def icon_url(self, size: int = eveimageserver._DEFAULT_IMAGE_SIZE) -> str:
        map_category_2_other = {
            self.CATEGORY_ALLIANCE: "alliance_logo_url",
            self.CATEGORY_CHARACTER: "character_portrait_url",
            self.CATEGORY_CORPORATION: "corporation_logo_url",
            self.CATEGORY_INVENTORY_TYPE: "type_icon_url",
        }
        if self.category not in map_category_2_other:
            return ""
        else:
            func = map_category_2_other[self.category]
            return getattr(eveimageserver, func)(self.entity_id, size=size)



from django.conf import settings

from app_utils.django import clean_setting

# switch to enable/disable ability to request standings for corporations
SR_CORPORATIONS_ENABLED = clean_setting("SR_CORPORATIONS_ENABLED", True)

# whether ESI requests have a timeout
SR_ESI_TIMEOUT_ENABLED = clean_setting("SR_ESI_TIMEOUT_ENABLED", True)

# Send notifications to users about the results of standings requests
SR_NOTIFICATIONS_ENABLED = clean_setting("SR_NOTIFICATIONS_ENABLED", True)

# Automatically sync standing for alt characters known to Auth
# that have standing in-game
SR_SYNC_BLUE_ALTS_ENABLED = clean_setting("SR_SYNC_BLUE_ALTS_ENABLED", True)

# Select the entity type of your standings master
SR_OPERATION_MODE = clean_setting(
    "SR_OPERATION_MODE", "alliance", choices=["alliance", "corporation"]
)

# This is a map, where the key is the State the user is in
# and the value is a list of required scopes to check
SR_REQUIRED_SCOPES = getattr(
    settings,
    "SR_REQUIRED_SCOPES",
    {"Member": ["publicData"], "Blue": [], "": []},  # no state
)

# Standing data will be considered stale and removed from the local
# database after the configured hours.
# The latest standings data will never be purged, no matter how old it is
SR_STANDINGS_STALE_HOURS = clean_setting("SR_STANDINGS_STALE_HOURS", 48)

# Max hours to wait for a standing to be effective after being marked actioned
# Non effective standing requests will be reset when this timeout expires.
SR_STANDING_TIMEOUT_HOURS = clean_setting("SR_STANDING_TIMEOUT_HOURS", 24)

# id of character to use for updating alliance contacts - needs to be set
STANDINGS_API_CHARID = clean_setting("STANDINGS_API_CHARID", None, required_type=int)

# charaters belonging to these alliances are considered to be "in organization"
STR_ALLIANCE_IDS = clean_setting("STR_ALLIANCE_IDS", None, required_type=list)

# charaters belonging to these corporations are considered to be "in organization"
STR_CORP_IDS = clean_setting("STR_CORP_IDS", None, required_type=list)



# Generated by Django 2.2.13 on 2020-07-05 15:47

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("standingsrequests", "0003_rename_non_pep"),
    ]

    operations = [
        migrations.AddField(
            model_name="evenamecache",
            name="category",
            field=models.CharField(
                choices=[
                    ("alliance", "alliance"),
                    ("character", "character"),
                    ("constellation", "constellation"),
                    ("corporation", "corporation"),
                    ("faction", "faction"),
                    ("inventory_type", "inventory_type"),
                    ("region", "region"),
                    ("solar_system", "solar_system"),
                    ("station", "station"),
                ],
                default=None,
                max_length=16,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="abstractstandingsrequest",
            name="action_by",
            field=models.ForeignKey(
                default=None,
                help_text="standing manager that accepted or rejected this requests",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="abstractstandingsrequest",
            name="action_date",
            field=models.DateTimeField(
                help_text="datetime of action by standing manager", null=True
            ),
        ),
        migrations.AlterField(
            model_name="abstractstandingsrequest",
            name="contact_id",
            field=models.PositiveIntegerField(
                db_index=True, help_text="EVE Online ID of contact this standing is for"
            ),
        ),
        migrations.AlterField(
            model_name="abstractstandingsrequest",
            name="contact_type_id",
            field=models.PositiveIntegerField(
                db_index=True, help_text="EVE Online Type ID of this contact"
            ),
        ),
        migrations.AlterField(
            model_name="abstractstandingsrequest",
            name="effective_date",
            field=models.DateTimeField(
                help_text="Datetime when this standing was set active in-game",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="abstractstandingsrequest",
            name="is_effective",
            field=models.BooleanField(
                default=False,
                help_text="True, when this standing is also set in-game, else False",
            ),
        ),
        migrations.AlterField(
            model_name="abstractstandingsrequest",
            name="request_date",
            field=models.DateTimeField(
                auto_now_add=True,
                db_index=True,
                help_text="datetime this request was created",
            ),
        ),
        migrations.AlterField(
            model_name="alliancestanding",
            name="contact_id",
            field=models.PositiveIntegerField(db_index=True),
        ),
        migrations.AlterField(
            model_name="alliancestanding",
            name="standing",
            field=models.FloatField(db_index=True),
        ),
        migrations.AlterField(
            model_name="characterassociation",
            name="alliance_id",
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AlterField(
            model_name="characterassociation",
            name="character_id",
            field=models.PositiveIntegerField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="characterassociation",
            name="corporation_id",
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AlterField(
            model_name="characterassociation",
            name="main_character_id",
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AlterField(
            model_name="contactlabel",
            name="label_id",
            field=models.BigIntegerField(db_index=True),
        ),
        migrations.AlterField(
            model_name="contactset",
            name="date",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="corpstanding",
            name="contact_id",
            field=models.PositiveIntegerField(db_index=True),
        ),
        migrations.AlterField(
            model_name="corpstanding",
            name="standing",
            field=models.FloatField(db_index=True),
        ),
        migrations.AlterField(
            model_name="evenamecache",
            name="entity_id",
            field=models.PositiveIntegerField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="evenamecache",
            name="updated",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="pilotstanding",
            name="contact_id",
            field=models.PositiveIntegerField(db_index=True),
        ),
        migrations.AlterField(
            model_name="pilotstanding",
            name="standing",
            field=models.FloatField(db_index=True),
        ),
    ]



import logging

from django.utils.translation import gettext_lazy as _

from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, ServicesHook

from . import __title__
from .models import StandingRequest, StandingRevocation
from .urls import urlpatterns

logger = logging.getLogger(__name__)


class StandingsRequestService(ServicesHook):
    def __init__(self):
        ServicesHook.__init__(self)
        self.name = "standingsrequests"
        self.urlpatterns = urlpatterns
        self.access_perm = StandingRequest.REQUEST_PERMISSION_NAME

    def delete_user(self, user, notify_user=False):
        logger.debug("Deleting user %s standings", user)
        StandingRequest.objects.delete_for_user(user)

    def validate_user(self, user):
        logger.debug("Validating user %s standings", user)
        if not self.service_active_for_user(user):
            self.delete_user(user)

    def service_active_for_user(self, user):
        return user.has_perm(self.access_perm)


@hooks.register("services_hook")
def register_service():
    return StandingsRequestService()


class StandingsRequestMenuItem(MenuItemHook):
    def __init__(self):
        MenuItemHook.__init__(
            self,
            _(__title__),
            "fas fa-plus-square fa-fw grayiconecolor",
            "standingsrequests:index",
        )

    def render(self, request):
        if request.user.has_perm("standingsrequests.affect_standings"):
            app_count = (
                StandingRequest.objects.pending_requests().count()
                + StandingRevocation.objects.pending_requests().count()
            )
            self.count = app_count if app_count and app_count > 0 else None
        if request.user.has_perm(StandingRequest.REQUEST_PERMISSION_NAME):
            return MenuItemHook.render(self, request)
        return ""


@hooks.register("menu_item_hook")
def register_menu():
    return StandingsRequestMenuItem()
